;;;; crunch - expander


(module (crunch compiler expand)

  (crunch-expand expanded-modules exported register-exports)

  (import (scheme base)
          (scheme cxr)
          chicken.base
          chicken.keyword
          chicken.syntax
          chicken.type
          chicken.format
          chicken.internal
          chicken.pretty-print
          srfi-1
          matchable
          (scheme file))

  (import (crunch compiler node)
          (crunch compiler util))

  (define expanded-modules (make-parameter '()))
  (define register-exports (make-parameter #t))
  (define exported (make-parameter '()))

  ;; marks exported identifiers - called by ##sys#register-export
  ;; note: guarded with parameter to avoid marking during meta-evaluation
  (set! ##sys#toplevel-definition-hook
    (lambda (sym renamed exported?)
      (when (and (register-exports)
                 (or exported? (eq? (register-exports) 'all))
                 (or (eq? 'main sym) (not (whole-program))))
        (exported (alist-cons sym renamed (exported))))))

  ;; we need this too in case module is false
  (set! ##sys#register-export
    (let ((old ##sys#register-export))
      (lambda (sym mod)
        (if (not mod) ; otherwise the hook above fires
            (let ((reg (register-exports)))
              (when (or (eq? reg 'all)
                        (eq? 'main sym)
                        (not (whole-program)))
                (exported (alist-cons sym sym (exported)))))
            (old sym mod)))))

  (define (compile-to-node exp env #!optional cntr evalenv tl?)

    (define-syntax thread-id
      (syntax-rules ()
        ((_ t) (##sys#slot t 14))))

    (define (find-id id se)         ; ignores macro bindings
      (cond ((null? se) #f)
            ((and (eq? id (caar se)) (symbol? (cdar se))) (cdar se))
            (else (find-id id (cdr se)))))

    (define (rename var)
      (cond ((find-id var (##sys#current-environment)))
            ((##sys#get var '##core#macro-alias) symbol? => values)
            (else var)))

    (define (lookup var0 e)
      (let ((var (rename var0)))
        (cond ((memq var e) (values var var0))
              (else (values #f var)))))

    (define (emit-syntax-trace-info tf info cntr)
      (when tf
        (##core#inline "C_emit_trace_info"
         (or (get-line-number info) "<syntax>")
         info
         cntr
         (thread-id ##sys#current-thread) ) ) )

    (define (genalias _) (gensym 'v))

    (define (handle-expansion-result outer-ln)
      (lambda (input output)
        (and-let* (((not (eq? input output)))
                   (ln (or (get-line-number input) outer-ln)))
          (##sys#update-line-number-database! output ln))
        output))

    (define (include-file x ci e tf cntr tl?)
      (##sys#include-forms-from-file
       (cadr x) (caddr x) ci
       (lambda (forms)
         (compile (if (pair? (cdddr x)) ; body?
                      (##sys#canonicalize-body (append forms (cadddr x))
                       (##sys#current-environment))
                      `(##core#begin ,@forms))
                  e #f tf cntr tl?))))

    (define (loc ln cntr)
      (or ln cntr))

    (define (compile x e h tf cntr tl?)
      (cond ((keyword? x) (node* cntr 'const (list x)))
            ((symbol? x)
             (let-values (((f v) (lookup x e)))
               (if (not f)
                   (let ((var (cond ((not (symbol? v)) x) ; syntax?
                                    ((assq x (##sys#current-environment)) v)
                                    (else (##sys#alias-global-hook v #f cntr)))))
                     (node* cntr 'global-ref
                       (list var (or (not (##sys#current-module))
                                             (assq var (exported)))
                                             x)))
                   (node* cntr 'local-ref (list f x)))))
            ((or (number? x)
                 (boolean? x)
                 (char? x)
                 (eof-object? x)
                 (bwp-object? x)
                 (string? x)
                 (bytevector? x)
                 (vector? x)
                 (##sys#srfi-4-vector? x))
             (node* cntr 'const (list x)))
            ((not (pair? x))
             (##sys#syntax-error/context "illegal non-atomic object" x))
            ((symbol? (car x))
             (emit-syntax-trace-info tf x cntr)
             (let* ((ln (get-line-number x))
                    (x2 (fluid-let ((chicken.syntax#expansion-result-hook
                                      (handle-expansion-result ln)))
                          (expand x (##sys#current-environment)))))
               (if (not (eq? x2 x))
                   (compile x2 e h tf cntr tl?)
                   (let ((head (rename (car x))))
                     ;; here we did't resolve ##core#primitive, but that is done in compile-call (via
                     ;; a normal walking of the operator)
                     (case head

                       ((##core#quote)
                        (let* ((c (strip-syntax (cadr x))))
                          (node* (loc ln cntr) 'const (list c))))

                       ((##core#syntax)
                        (let ((c (cadr x)))
                          (node* (loc ln cntr) 'const (list c))))

                       ((##core#check)
                        (compile (cadr x) e h tf cntr #f) )

                       ((##core#immutable)
                        (compile (cadr x) e h tf cntr #f) )

                       ((##core#undefined) (node* (loc ln cntr) 'unspecified))

                       ((##core#if)
                        (let* ((test (compile (cadr x) e #f tf cntr #f))
                               (cns (compile (caddr x) e #f tf cntr #f))
                               (alt (if (pair? (cdddr x))
                                        (compile (cadddr x) e #f tf cntr #f)
                                        (compile '(##core#undefined) e #f tf cntr #f) ) ) )
                          (node* (loc ln cntr) 'if '() test cns alt)))

                       ((##core#begin)
                        (let* ((body (cdr x))
                               (len (length body)) )
                          (case len
                            ((0) (compile '(##core#undefined) e #f tf cntr tl?))
                            ((1) (compile (car body) e #f tf cntr tl?))
                            (else (apply node* (loc ln cntr) 'begin '()
                                         (map (lambda (x) (compile x e #f tf cntr tl?)) body))))))

                       ((##core#ensure-toplevel-definition)
                        (unless tl?
                          (error "toplevel definition in non-toplevel context for variable" (cadr x)))
                        (##sys#put/restore! (cadr x) '##sys#override 'value)
                        (compile '(##core#undefined) e #f tf cntr #f))

                       ((##core#set!)
                        (let ((var (cadr x)))
                          (receive (i j) (lookup var e)
                            (let ((val (compile (caddr x) e var tf cntr #f)))
                              (cond ((not i)
                                     (and-let* ((a (assq var (##sys#current-environment)))
                                                ((symbol? (cdr a))))
                                       (error "assignment to imported value binding" var))
                                     (let ((qvar (##sys#alias-global-hook j #t cntr)))
                                       (node* (loc ln cntr) 'global-set!
                                              (list qvar
                                                    (or (assq var (exported))
                                                        (if (whole-program)
                                                            (eq? var 'main)
                                                            (not (##sys#current-module))))
                                                    var)
                                              val)))
                                    (else (node* (loc ln cntr) 'local-set! (list i var) val)))))))

                       ((##core#let)
                        (let* ((bindings (cadr x))
                               (n (length bindings))
                               (vars (map car bindings))
                               (aliases (map genalias vars))
                               (e2 (append aliases e))
                               (se2 (##sys#extend-se (##sys#current-environment) vars aliases))
                               (body (parameterize ((##sys#current-environment se2))
                                       (compile-to-node
                                         (##sys#canonicalize-body (cddr x) (##sys#current-environment) #f)
                                         e2 cntr evalenv #f)) ) )
                          (apply node* (loc ln cntr) 'let (list vars aliases)
                                (append (map (lambda (b)
                                               (compile (cadr b) e (car b) tf cntr #f))
                                             bindings)
                                        (list body)))))

                       ((##core#letrec*)
                        (let ((bindings (cadr x))
                              (body (cddr x)) )
                          (compile
                            `(##core#let
                              ,(##sys#map (lambda (b)
                                            (list (car b) '(##core#undefined)))
                                bindings)
                              ,@(##sys#map (lambda (b)
                                             `(##core#set! ,(car b) ,(cadr b)))
                                 bindings)
                              (##core#let () ,@body) )
                            e h tf cntr #f)))

                       ((##core#letrec)
                        (let* ((bindings (cadr x))
                               (vars (map car bindings))
                               (tmps (map gensym vars))
                               (body (cddr x)) )
                          (compile
                            `(##core#let
                              ,(map (lambda (b)
                                      (list (car b) '(##core#undefined)))
                                 bindings)
                              (##core#let ,(map (lambda (t b) (list t (cadr b))) tmps bindings)
                               ,@(map (lambda (v t)
                                        `(##core#set! ,v ,t))
                                   vars tmps)
                               (##core#let () ,@body) ) )
                            e h tf cntr #f)))

                       ((##core#lambda)
                        (##sys#check-syntax
                          'lambda x '(_ lambda-list . #(_ 1)) #f
                          (##sys#current-environment))
                        (let loop ((llist (cadr x))
                                   (argc 0)
                                   (vars '())
                                   (opt #f))
                          (cond ((null? llist)
                                  (let* ((vars (reverse vars))
                                         (aliases (map genalias vars))
                                         (e2 (append aliases e))
                                         (se2 (##sys#extend-se (##sys#current-environment) vars aliases))
                                         (body
                                           (parameterize ((##sys#current-environment se2))
                                             (compile-to-node
                                               (##sys#canonicalize-body (cddr x) se2 #f)
                                               e2 (or h cntr) evalenv #f)) ) )
                                    (node*
                                      (loc ln cntr)
                                      'lambda
                                      (list (gensym 'f) argc vars aliases h)
                                      body)))
                                ((symbol? llist)
                                  (if (zero? argc)
                                      (loop '() 'values (list llist) #f)
                                          (error "ill-formed lambda list" (cadr x))))
                                ((not (pair? llist))
                                  (error "ill-formed lambda list" (cadr x)))
                                ((eq? '#!optional (car llist))
                                  (loop (cdr llist) argc vars #t))
                                ((symbol? (car llist))
                                  (loop (cdr llist)
                                            (if opt argc (add1 argc))
                                            (cons (car llist) vars)
                                            opt))
                                (else (error "ill-formed lambda list" (cadr x))))))

                     ((##core#let-syntax)
                      (parameterize
                        ((##sys#current-environment
                          (append
                            (map (lambda (b)
                                   (list
                                     (car b)
                                     (##sys#current-environment)
                                     (##sys#ensure-transformer
                                      (meta-eval (cadr b))
                                      (strip-syntax (car b)))))
                              (cadr x) )
                            (##sys#current-environment)) ) )
                        (compile
                          (##sys#canonicalize-body (cddr x) (##sys#current-environment) #f)
                          e #f tf cntr #f)))

                     ((##core#letrec-syntax)
                      (let* ((ms (map (lambda (b)
                                        (list
                                          (car b)
                                          #f
                                          (##sys#ensure-transformer
                                           (meta-eval (cadr b))
                                           (strip-syntax (car b)))))
                                   (cadr x) ) )
                             (se2 (append ms (##sys#current-environment))) )
                        (for-each
                          (lambda (sb)
                            (set-car! (cdr sb) se2) )
                          ms)
                        (parameterize ((##sys#current-environment se2))
                          (compile
                            (##sys#canonicalize-body (cddr x) (##sys#current-environment) #f)
                            e #f tf cntr #f))))

                     ((##core#define-syntax)
                      (let* ((var (cadr x))
                             (body (caddr x))
                             (name (rename var)))
                        (##sys#put/restore! name '##sys#override 'syntax)
                        (##sys#register-syntax-export
                         name (##sys#current-module)
                         body) ; not really necessary, it only shouldn't be #f
                        (##sys#extend-macro-environment
                         name
                         (##sys#current-environment)
                         (meta-eval body))
                        (compile '(##core#undefined) e #f tf cntr #f) ) )

                     ((##core#define-compiler-syntax)
                      (compile '(##core#undefined) e #f tf cntr #f))

                     ((##core#let-compiler-syntax)
                      (compile
                        (##sys#canonicalize-body (cddr x) (##sys#current-environment) #f)
                        e #f tf cntr #f))

                     ((##core#include)
                      (include-file x #f e tf cntr tl?))

                     ((##core#include-ci)
                      (include-file x #t e tf cntr tl?))

                     ((##core#let-module-alias)
                      (##sys#with-module-aliases
                       (map (lambda (b)
                              (##sys#check-syntax 'functor b '(symbol symbol))
                              (strip-syntax b))
                         (cadr x))
                       (lambda ()
                         (compile `(##core#begin ,@(cddr x)) e #f tf cntr tl?))))

                     ((##core#module)
                      (let* ((x (strip-syntax x))
                             (name (cadr x))
                             (exports
                               (or (eq? #t (caddr x))
                                   (map (lambda (exp)
                                          (cond ((symbol? exp) exp)
                                                ((and (pair? exp)
                                                      (let loop ((exp exp))
                                                        (or (null? exp)
                                                            (and (symbol? (car exp))
                                                                 (loop (cdr exp))))))
                                                 exp)
                                                (else
                                                  (##sys#syntax-error
                                                   'module
                                                   "invalid export syntax" exp name))))
                                     (caddr x)))))
                        (when (##sys#current-module)
                          (##sys#syntax-error 'module "modules may not be nested" name))
                        (parameterize ((##sys#current-module
                                        (##sys#register-module name #f exports))
                                       (##sys#current-environment '())
                                       (exported '())
                                       (##sys#macro-environment
                                        ##sys#initial-macro-environment)
                                       (##sys#module-alias-environment
                                        (##sys#module-alias-environment)))
                          (##sys#with-property-restore
                           (lambda ()
                             (let loop ((body (cdddr x)) (xs '()))
                               (if (null? body)
                                   (let ((xs (reverse xs)))
                                     (##sys#finalize-module (##sys#current-module))
                                     (expanded-modules (cons (##sys#current-module)
                                                             (expanded-modules)))
                                    (##sys#provide (module-requirement name))
                                     (apply node* ln 'begin '() xs))
                                   (loop
                                     (cdr body)
                                     (cons (compile
                                             (car body)
                                             '() #f tf cntr
                                             #t) ; reset back to toplevel!
                                           xs))))) ) )))

                     ((##core#loop-lambda)
                      (compile `(, (rename 'lambda) ,@(cdr x)) e #f tf cntr #f) )

                     ((##core#c-lambda)
                      (node* ln 'c-lambda (cons (gensym 'f) (strip-syntax (cdr x)))))

                     ((##core#c-external)
                      (match (cdr x)
                        ((inc name args ret)
                          (node* ln 'c-external
                            (list inc (let ((name (strip-syntax name)))
                                        (if (string? name) (string->symbol name) name))
                                      (strip-syntax args)
                                      (strip-syntax ret))))))

                     ((##core#c-declare)
                      (node* ln 'c-declare (cdr x)))

                     ((##core#provide)
                      (compile `(##sys#provide (##core#quote ,(cadr x))) e #f tf cntr #f))

                     ((##core#require-for-syntax)
                      (chicken.load#load-extension (cadr x) #f #f)
                      (compile '(##core#undefined) e #f tf cntr #f))

                     ((##core#require)
                      (compile '(##core#undefined) e #f tf cntr #f))

                     ((##core#elaborationtimeonly ##core#elaborationtimetoo) ; <- Note this!
                      (meta-eval (cadr x))
                      (compile '(##core#undefined) e #f tf cntr tl?) )

                     ((##core#compiletimetoo)
                      (meta-eval (cadr x))
                      (compile (cadr x) e #f tf cntr tl?) )

                     ((##core#compiletimeonly ##core#callunit ##core#local-specialization ##core#declare)
                      (compile '(##core#undefined) e #f tf cntr tl?) )

                     ((##core#define-inline ##core#define-constant)
                      (compile `(, (rename 'define) ,@(cdr x)) e #f tf cntr tl?) )

                     ((##core#primitive ##core#inline ##core#inline_allocate ##core#foreign-lambda
                       ##core#define-foreign-variable
                       ##core#define-external-variable ##core#let-location
                       ##core#foreign-primitive ##core#location
                       ##core#foreign-lambda* ##core#define-foreign-type)
                      (##sys#syntax-error "cannot compile compiler-special-form" x) )

                     ((##core#app)
                      (compile-call (cdr x) e tf cntr (##sys#current-environment)) )

                     ((##core#the)
                      (compile (cadddr x) e h tf cntr tl?))

                     ;;((##core#typecase) XXX

                     ((##core#typecase)
                      ;; drops exp and requires "else" clause
                      (cond ((assq 'else (strip-syntax (cdddr x))) =>
                             (lambda (cl)
                               (compile (cadr cl) e h tf cntr tl?)))
                            (else
                              (##sys#syntax-error
                               'compiler-typecase
                               "no `else-clause' in unresolved `compiler-typecase' form"
                               x))))

                     (else
                       (fluid-let ((##sys#syntax-context (cons head ##sys#syntax-context)))
                         (compile-call x e tf cntr (##sys#current-environment)))))))))

      (else
        (emit-syntax-trace-info tf x cntr)
        (compile-call x e tf cntr (##sys#current-environment))) ) )

  (define (meta-eval x)
    (parameterize ((register-exports #f))
      (##sys#eval/meta x)))

  (define (fudge-argument-list n alst)
    (if (null? alst)
        (list alst)
        (do ((n n (- n 1))
             (c 0 (+ c 1))
             (args alst
                   (if (eq? '() args)
                       (error "bad argument count" n c)
                       (cdr args)))
             (last #f args) )
            ((= n 0)
             (set-cdr! last (list args))
             alst) ) ) )

  (define (checked-length lst)
    (let loop ((lst lst) (n 0))
      (cond ((null? lst) n)
            ((pair? lst) (loop (cdr lst) (+ n 1)))
            (else #f) ) ) )

  (define (compile-call x e tf cntr se)
    (let* ((head (car x))
           (fn (compile (car x) e #f tf cntr #f))
           (args (cdr x))
           (argc (checked-length args))
           (info x)
           (ln (or (get-line-number info)
                   (or cntr "toplevel call"))))
      (if argc
          (apply node* ln 'call '(#f) fn (map (lambda (arg) (compile arg e #f tf cntr #f)) args))
          (##sys#syntax-error/context "malformed expression" x))))

  (compile exp env #f #f cntr tl?) )

(define (crunch-expand x #!optional env)
  (fluid-let ((##sys#features (cons #:crunched ##sys#features)))
    (compile-to-node x '() #f #f #t)))

)



